/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "common.h" 
#include "support.h"
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>

#define LOW_NIBBLE 0x0F 

artnet_node node ;
artnet_node_entry active_node ;
int nodes_found ;
int transmiting = 0 ;

int8_t output_data[ARTNET_MAX_PORTS][ARTNET_MAX_DMX] ;

void display_packet_data( GtkWidget *data, artnet_packet p, const gchar *bstr, const gchar *header, const gchar *extra, const gchar *fmt, ... ) {
	va_list ap;
	gchar *str, *header_str ;
	GtkTextView *text;
	GtkTextBuffer *buffer ;
	GtkTextIter iter;
	GtkCheckButton *button ;
	GtkCheckButton *header_button ;
	
	text = (GtkTextView*) lookup_widget((GtkWidget *) data, "textview1") ;
	buffer = gtk_text_view_get_buffer((GtkTextView*) text) ;
	button = (GtkCheckButton*) lookup_widget((GtkWidget *) data, bstr) ;
	header_button = (GtkCheckButton*) lookup_widget((GtkWidget *) data, "header_button") ;
	
	if(! gtk_toggle_button_get_active((GtkToggleButton*) button)) {
		va_start(ap, fmt) ;

		str = g_strdup_vprintf(fmt, ap) ;
		header_str = g_strdup_printf("%s: [%s]\n", inet_ntoa(p->from), header) ;
		
		gtk_text_buffer_get_end_iter(buffer, &iter) ;
		gtk_text_buffer_insert(buffer,&iter, header_str , -1) ;
		
		if(! gtk_toggle_button_get_active((GtkToggleButton*) header_button)) {
			gtk_text_buffer_get_end_iter(buffer, &iter) ;
			gtk_text_buffer_insert(buffer,&iter, str , -1) ;

			if(extra != NULL) {
				gtk_text_buffer_get_end_iter(buffer, &iter) ;
				gtk_text_buffer_insert(buffer,&iter, extra , -1) ;			
			}
			gtk_text_buffer_get_end_iter(buffer, &iter) ;
			gtk_text_buffer_insert(buffer,&iter, "------------------------------------------------------------------------------------------------------------------------\n" , -1) ;

			gtk_text_buffer_get_end_iter(buffer, &iter) ;
			gtk_text_view_scroll_to_iter(text,&iter, 0.0,FALSE, 0.0, 1.0) ;
			free(str); 

			va_end(ap) ;
		}
	}
}

void display_packet(artnet_packet p, GtkWidget* data) {
	gchar *str ;
	int i;
	
	switch(p->type) {
		case ARTNET_POLL:
				display_packet_data(data, p, "ignore_ap_button", "ArtPoll", NULL,
					"{0x%04hx //OpCode}, {%d //Prot Vers Hi}, {%d //Prot Vers Lo},\n{0x%02hhx //TalkToMe}, {%d //Pad} }\n",
					p->data.ap.opCode, p->data.ap.verH, p->data.ap.ver, p->data.ap.ttm, p->data.ap.pad) ;
			break;
		case ARTNET_REPLY:
		   		display_packet_data(data, p, "ignore_ar_button", "ArtPollReply", NULL,
					"{ {%s //ID}, {0x%04hx //OpCode},\n{ %d.%d.%d.%d:0x%04hx //Node IP:Port}, {0x%02hhx%02hhx //Firmware Version}, {0x%02hhx //Sub-Net Switch hi}\n{0x%02hhx //Sub-Net Switch lo}, {0x%02hhx%02hhx //Oem code}, {%d //UBEA Version}, {0x%02hhx //Status bits},\n{0x%02hhx%02hhx ///Esta Manufacturer Code},\n {%s //ShortName}, {%s //LongName},\n{%s //NodeReport},\n{%d //Number Ports Lo}, {[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //PortTypes},\n{[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //GoodInput}, {[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //GoodOutput},\n {[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //Swin}, {[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //Swout},\n{0x%02hhx //Sw Video}, {0x%02hhx //Sw Macro}, {0x%02hhx //Sw Remote},\n{%02d-%02d-%02d-%02d-%02d-%02d-%02d-%02d //MAC Address} }\n",
					p->data.ar.id, p->data.ar.opCode, p->data.ar.ip[0], p->data.ar.ip[1], p->data.ar.ip[2], p->data.ar.ip[3], p->data.ar.port, p->data.ar.verH, p->data.ar.ver, p->data.ar.subH, p->data.ar.sub, p->data.ar.oemH, p->data.ar.oem, p->data.ar.ubea, p->data.ar.status, p->data.ar.etsaman[0], p->data.ar.etsaman[1], p->data.ar.shortname, p->data.ar.longname, p->data.ar.nodereport, p->data.ar.numbports, p->data.ar.porttypes[0], p->data.ar.porttypes[1], p->data.ar.porttypes[2], p->data.ar.porttypes[3], p->data.ar.goodinput[0], p->data.ar.goodinput[1], p->data.ar.goodinput[2], p->data.ar.goodinput[3], p->data.ar.goodoutput[0], p->data.ar.goodoutput[1], p->data.ar.goodoutput[2], p->data.ar.goodoutput[3], p->data.ar.swin[0], p->data.ar.swin[1], p->data.ar.swin[2], p->data.ar.swin[3], p->data.ar.swout[0], p->data.ar.swout[1], p->data.ar.swout[2], p->data.ar.swout[3], p->data.ar.swvideo, p->data.ar.swmacro, p->data.ar.swremote, p->data.ar.mac[0], p->data.ar.mac[1], p->data.ar.mac[2], p->data.ar.mac[3], p->data.ar.mac[4], p->data.ar.mac[5], p->data.ar.mac[6], p->data.ar.mac[7] ) ;
			break;
		case ARTNET_INPUT:
		   		display_packet_data(data, p, "ignore_ai_button", "ArtInput", NULL,
					"{ {0x%04hx //OpCode}, {%d //Prot Vers Hi}, {%d //Prot Vers Lo}, {%d //Filler 1}, {%d //Filler 2},\n{%d //Number Ports},\n{[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //Input enables} }\n",
					p->data.ainput.opCode, p->data.ainput.verH, p->data.ainput.ver, p->data.ainput.filler1, p->data.ainput.filler2, p->data.ainput.numbports, p->data.ainput.input[0], p->data.ainput.input[1], p->data.ainput.input[2], p->data.ainput.input[3] );
			break;
		case ARTNET_ADDRESS:
		   		display_packet_data(data, p, "ignore_aa_button", "ArtAddress", NULL,
					"{ {0x%04hx //OpCode}, {%d //Prot Vers Hi}, {%d //Prot Vers Lo}, {%d //Filler 1}, {%d //Filler 2},\n{%s //ShortName}, {%s //LongName},\n{[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //Swin}, {[0x%02hhx] [0x%02hhx] [0x%02hhx] [0x%02hhx] //Swout}\n{0x%02hhx //Subnet}, {0x%02hhx //SwVideo}, {0x%02hhx //Command} }\n",
					p->data.addr.opCode, p->data.addr.verH, p->data.addr.ver, p->data.addr.filler1, p->data.addr.filler2, p->data.addr.shortname, p->data.addr.longname, p->data.addr.swin[0], p->data.addr.swin[1], p->data.addr.swin[2], p->data.addr.swin[3], p->data.addr.swin[0], p->data.addr.swin[1], p->data.addr.swin[2], p->data.addr.swin[3], p->data.addr.subnet, p->data.addr.command );
			break;
		case ARTNET_DMX:
		   		display_packet_data(data, p, "ignore_dmx_button", "ArtDMX", NULL,
					"{ {0x%04hx //OpCode}, {%d //Prot Vers Hi}, {%d //Prot Vers Lo},\n{0x%02hhx //Sequence}, {%d //Physical}, {0x%02hhx //Universe}, {%d //Channel Count},\n{Ch0=%02hhx, Ch1=%02hhx, Ch2=%02hhx, Ch3=%02hhx, Ch4=%02hhx, Ch5=%02hhx, Ch6=%02hhx, Ch7=%02hhx, Ch8=%02hhx, ... Ch512=%02hhx} }\n",
					p->data.admx.opCode, p->data.admx.verH, p->data.admx.ver, p->data.admx.sequence, p->data.admx.physical, p->data.admx.universe, p->data.admx.length, p->data.admx.data[0], p->data.admx.data[1], p->data.admx.data[2], p->data.admx.data[3], p->data.admx.data[4], p->data.admx.data[5], p->data.admx.data[6], p->data.admx.data[7], p->data.admx.data[8], p->data.admx.data[511] );
			break;
		case ARTNET_TODREQUEST:
				str = g_strdup("{ ") ;
				for( i=0 ; i < p->data.todreq.adCount; i++) {
					str = g_strconcat(str, g_strdup_printf("[%02hhx], ", p->data.todreq.address[i]), NULL ) ;
				}
				str = g_strconcat(str, " } //Universe Array )\n", NULL );
				display_packet_data(data, p, "ignore_tod_request_button", "ArtTodRequest", str,
					"{ {0x%04hx //OpCode}, {%d //Prot Vers Hi}, {%d //Prot Vers Lo},\n{0x%02hhx, 0x%02hhx //Fillers}, {0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx //Spares},\n{0x%02hhx //Command = %s}, {0x%02hhx //AdCount},\n",
					p->data.todreq.opCode, p->data.todreq.verH, p->data.todreq.ver, p->data.todreq.filler1, p->data.todreq.filler2, p->data.todreq.spare1, p->data.todreq.spare2, p->data.todreq.spare3, p->data.todreq.spare4, p->data.todreq.spare5, p->data.todreq.spare6, p->data.todreq.spare7, p->data.todreq.spare8, p->data.todreq.command,  p->data.todreq.command == ARTNET_TOD_FULL ? "FULL" : "NAK", p->data.todreq.adCount );
				
				free(str); 
			break;
		case ARTNET_TODDATA:
			
				display_packet_data(data, p, "ignore_tod_data_button", "ArtTodData", NULL,
					"{ {0x%04hx //OpCode}, {%d //Prot Vers Hi}, {%d //Prot Vers Lo},\n{0x%02hhx //Filler}, {0x%02hhx //Port},\n{0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx //Spares},\n{0x%02hhx //Command Response= %s}, {0x%02hhx //Address},\n",
					p->data.toddata.opCode, p->data.toddata.verH, p->data.toddata.ver, 
					p->data.toddata.filler1, p->data.toddata.port,
					p->data.toddata.spare1, p->data.toddata.spare2, p->data.toddata.spare3, p->data.toddata.spare4, p->data.toddata.spare5, p->data.toddata.spare6, p->data.todreq.spare7, p->data.toddata.spare8,
					p->data.toddata.cmdRes, p->data.toddata.cmdRes == ARTNET_TOD_FULL ? "FULL" : "NAK", p->data.toddata.address );
				
		
		
	   default :
		printf("unknown\n") ;
	   		break;
	};
	
}


int on_packet_recv(artnet_node n, void *pp, void *data) {
	display_packet((artnet_packet) pp, (GtkWidget*) data) ;
	return 0;
}


int on_packet_send(artnet_node n, void *pp, void *data) {
	display_packet((artnet_packet) pp, (GtkWidget*) data) ;
	return 0;
}

int on_poll_recv(artnet_node n, void *pp, void *data) {
	GtkCheckButton *button ;
	button = (GtkCheckButton*) lookup_widget((GtkWidget *) data, "autopollreply_checkbutton") ;
	
	if(! gtk_toggle_button_get_active((GtkToggleButton*) button)) 
		return 1;
	else
		return 0 ;
	
}

int on_tod_request_recv(artnet_node n, void *pp, void *data) {
	GtkCheckButton *button ;
	button = (GtkCheckButton*) lookup_widget((GtkWidget *) data, "autotodrequest_reply_checkbutton") ;
	
	if(! gtk_toggle_button_get_active((GtkToggleButton*) button)) 
		return 1;
	else
		return 0 ;
	
}



artnet_node_entry lookup_node_entry_by_id(artnet_node n, int id) {
	artnet_node_list nl = artnet_get_nl(n) ;
	artnet_node_entry e = artnet_nl_first(nl) ;
	int i = 1;
	
	if(id < 1 || id > MAX_BUTTONS)
		return NULL;
	
	
	while(e != NULL && i <= MAX_BUTTONS) {
		if(i == id)
			break;

		e = artnet_nl_next(nl) ;
		i++ ;
	}
	
	if(i == id && e != NULL) 
		return e ;

	return NULL ;
	
	
}

// this is probably not a good way to go, everytime we get a poll_callback
// we update the gui.
int on_reply_recv(artnet_node n, void *pp, void *data) {
	artnet_node_list nl = artnet_get_nl(n) ;
	
	artnet_node_entry e = artnet_nl_first(nl) ;
	GtkButton *button ;
	GtkRadioButton *rbutton, *tbutton, *rembutton ;
	gchar *gstr ;
	
	if(nodes_found != artnet_nl_get_length(nl)) {
		nodes_found++ ;
		printf("updating gui for node %d\n", nodes_found) ;
		
		e = lookup_node_entry_by_id(n, nodes_found) ;
			
		gstr = g_strdup_printf("nodebutton%d", nodes_found) ;
		button = (GtkButton*) lookup_widget((GtkWidget *) data, gstr ) ;
		gstr = g_strdup_printf("recvradiobutton%d", nodes_found) ;
		rbutton = (GtkRadioButton*) lookup_widget((GtkWidget *) data, gstr ) ;
		gstr = g_strdup_printf("transradiobutton%d", nodes_found) ;
		tbutton = (GtkRadioButton*) lookup_widget((GtkWidget *) data, gstr ) ;
		gstr = g_strdup_printf("remradiobutton%d", nodes_found) ;
		rembutton = (GtkRadioButton*) lookup_widget((GtkWidget *) data, gstr ) ;
		gstr = g_strdup_printf("%d.%d.%d.%d", e->ip[0], e->ip[1], e->ip[2], e->ip[3]) ;

		gtk_button_set_label(button, gstr) ;
		gtk_button_set_label((GtkButton*) rbutton, gstr) ;
		gtk_button_set_label((GtkButton*) tbutton, gstr) ;
		gtk_button_set_label((GtkButton*) rembutton, gstr) ;
		
		gtk_widget_set_sensitive((GtkWidget*) button, TRUE) ;
		gtk_widget_set_sensitive((GtkWidget*) rbutton, TRUE) ;
		gtk_widget_set_sensitive((GtkWidget*) tbutton, TRUE) ;
		gtk_widget_set_sensitive((GtkWidget*) rembutton, TRUE) ;
		
		if(active_node == NULL && artnet_nl_first(nl) != NULL)
			change_active_node_to(button, n, 1) ;
	}
	
	return 0;
}


void change_active_node_to(GtkButton *button, artnet_node n, int id) {
	artnet_node_entry e = lookup_node_entry_by_id(n, id) ;
	GtkEntry *shortentry, *longentry, *reportentry ;
	int i ;
	
	printf("  CHANGING ACTIVE NODE!\n") ;
	if(e != NULL) {
		active_node = e ;
		shortentry = (GtkEntry*) lookup_widget((GtkWidget *) button, "shortnameentry") ;
		longentry = (GtkEntry*) lookup_widget((GtkWidget *) button, "longnameentry") ;
		reportentry = (GtkEntry*) lookup_widget((GtkWidget *) button, "reportentry") ;
		
		// change this node to match our new peer
		// else we won't get any dmx packets
		printf("set subnet address\n") ;
		artnet_set_subnet_addr(n, e->sub) ;

		for(i=0; i < MIN(e->numbports, ARTNET_MAX_PORTS) ; i++) {
  			// set the universe address 
			printf("  setting input port address\n") ;
  			artnet_set_port_addr(node, i, ARTNET_INPUT_PORT, e->swout[i] & 0x0F) ;
  			artnet_set_port_addr(node, i, ARTNET_OUTPUT_PORT, e->swin[i] & 0x0F) ;

		}
		
		gtk_entry_set_text(shortentry, e->shortname) ;
		gtk_entry_set_text(longentry, e->longname) ;
		gtk_entry_set_text(reportentry, e->nodereport) ;
		
	}
}

void update_recv_page(GtkWidget *widget, artnet_node n) {
	int i;
	gchar *gstr ;
	GtkCheckButton *cbutton ;
	
	if (active_node == NULL)
		return ;
	
	for(i=0 ; i <ARTNET_MAX_PORTS;i++) {
		gstr = g_strdup_printf("recvcheckbutton%d", i+1) ;
		cbutton = (GtkCheckButton*) lookup_widget(widget, gstr) ;
		gstr = g_strdup_printf("Universe %d - %d", active_node->sub, active_node->swin[i] & LOW_NIBBLE) ;
		gtk_button_set_label((GtkButton*) cbutton, gstr) ;
		
	}
}

void update_trans_page(GtkWidget *widget, artnet_node n) {
	
	int i;
	gchar *gstr ;
	GtkLabel *label ;
	
	if (active_node == NULL)
		return ;
		
	for(i=0 ; i <ARTNET_MAX_PORTS;i++) {
		gstr = g_strdup_printf("translabel%d", i+1) ;
		label = (GtkLabel*) lookup_widget(widget, gstr) ;
		gstr = g_strdup_printf("Universe %d - %d", active_node->sub, active_node->swout[i] & LOW_NIBBLE) ;
		gtk_label_set_text((GtkLabel*) label, gstr) ;
		
	}
}


void update_remote_page(GtkWidget *widget, artnet_node n) {
	printf("updating remote page\n") ;
	if (active_node == NULL)
		return ;
		
	
}



int on_dmx_recv(artnet_node n, int port, void *d) {
	int length, i ;
	int8_t *data ;
	gchar *gstr ;
	GtkTreeView *tree;
	GtkListStore *store ;
	GtkTreeIter iter ;
	
	data = artnet_read_dmx(n, port, &length) ;
	
	gstr = g_strdup_printf("recv_treeview%i", port+1) ;
	tree = (GtkTreeView*) lookup_widget((GtkWidget*) d, gstr) ;
	store = (GtkListStore*) gtk_tree_view_get_model(tree) ;

	
	gtk_tree_model_get_iter_from_string((GtkTreeModel*) store, &iter, "0") ;

	for(i=0; i< length; i++) {
		gtk_list_store_set(store, &iter, COLUMN_DATA, *data, -1) ;
		data++ ;
		gtk_tree_model_iter_next((GtkTreeModel*) store, &iter) ;
	}


	return 0;
}


int tx_timeout(gpointer data) {
	GtkVScale *vscale ;
	gchar *gstr ;
	gdouble val ;
	int i ;
	
	if(transmiting == 0)
		return FALSE ;
	
	for(i=1; i<= ARTNET_MAX_PORTS; i++) {
		gstr = g_strdup_printf("trans_vscale%i", i) ;	
		vscale = (GtkVScale*) lookup_widget((GtkWidget*) data, gstr) ;
		val = gtk_range_get_value((GtkRange*) vscale) ;
		val++ ;
	
		if(val > 255)
			val = 0 ;
	
		gtk_range_set_value((GtkRange*) vscale, val) ;
		
		transmit(node, i-1, (GtkRange*) vscale) ;

	}
	
	return TRUE ;
	
}

void transmit(artnet_node n, int port, GtkRange *range) {
	int i;
	int8_t val ;

	val = gtk_range_get_value( range) ;
	
	for(i=0; i< ARTNET_MAX_DMX; i++) {
		if(i%2)
		  output_data[port][i] = val ;	
	    else
		  output_data[port][i] = 255 - val ;
	}
	
	artnet_send_dmx(node, port-1, ARTNET_MAX_DMX, &output_data[port][0]) ;
}
